{{

This is a derivation of 'jtc_c3_sdspiFemto.spin'.

It shaves sdspi down mostly to the low-level bus handling
happening in the cog, with hardware specific functions factored
out into higher level spin modules.

A core design goal is providing locked access to
the SPI bus and MUX on the C3, such that it is useable by multiple
cogs will compete for the lock which controls bus access.

All bus handling is via the assembly module, which reads/writes
via buffer pointers sent to the cog. Spin "drivers" provide the
command processing and "APIs" to developers.

A high level description of the changes relative to sdspi

- this code will only support C3 platform built-in hardware, and it's unique design.
- any form of booting code had been removed for now (it will return)
- the i2c code is mostly the same as sdspi
- the SPI code is almost entirely re-written, and made general (IE: not specific to SD cards)
- there are other design goals in mind, which will be revealed later, depending on the
  results of the dependancy of shrinking this code down even further.


Copyright (c) 2011 Joe Hohertz

----------------------------

Original description and credits here:

Modified Tiny Basic for use with Propeller Demo Board and Hydra.
I2C and SPI driver initialization & interface object derived from Propeller OS.

-Modified(hacked)SPIN routines in from Andre' Lamothe's C3SD_DEMO_010.SPIN
  and disabled ASM routines for minimal support for C3 boardn -JT Cook
Copyright (c) 2009 Michael Green.  See end of file for terms of use.
}}


CON

  ' SPI (serial peripheral interface) pins
  PIN_SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  PIN_SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  PIN_SPI_MOSI               = 9  ' SPI master out serial in to slave (spiDI)
  PIN_SPI_MISO               = 10  ' SPI master in serial out from slave (spiDO), 10 for good C3, change do 7 for BROKEN C3 with YELLOW jumper.
  PIN_SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.

  ' interface header SPI selects
  SPI_CHAN_SPI6          = 6 ' Header Interface SPI6.
  SPI_CHAN_SPI7          = 7 ' Header Interface SPI7.

'' Command code and error information for I2C driver
'' (For convenience in using just OS_loaderInit, these are included here.  The "master"
''  copies are considered the ones in OS_loader and these must be kept up-to-date).

  i2cReadCmd   = %00000001                      ' Read from EEPROM to HUB RAM (16 bit addresses)
  i2cWriteCmd  = %00000010                      ' Write to EEPROM from HUB RAM (16 bit addresses)
  i2cRead1Cmd  = %00000011                      ' Read from a device with only 8-bit addresses
  i2cWrite1Cmd = %00000100                      ' Write to a device with only 8-bit addresses
  i2cRead0Cmd  = %00000101                      ' Read from a device without address bytes
  i2cWrite0Cmd = %00000110                      ' Write to a device without address bytes
  i2cShutDown  = %00001000                      ' Stop I2C Bus (not yet implemented)
  i2cInit      = %00001001                      ' Initialize the I2C bus/states

  ioSpiInit      = %00010000                      ' Initialize the specified SPI bus and SD card
  ioSpiSelect    = %00010001                      ' Select a device on SPI or deselect
  ioSpiRead      = %00010010                      ' Read one or more bytes from selected SPI device
  ioSpiWrite     = %00010011                      ' Write one or more bytes from the selected SPI device
  ioSpiWriteRead = %00010100                      ' Read, and write, one or more bytes from SPI
  ioSpiWRBits    = %00010101                       ' Read, and write, 1-32 bits from SPI
  ioSpiReadBits  = %00010110                      ' read 1-32 bits
  ioSpiWriteBits = %00010111                      ' write 1-32 bits
  ioSpiShutdown  = %00011000                      ' Change all SD card pins to inputs

  ioCmdMask   = %00011111                      ' Used to mask off command bits
  spiCmdMask  = %00010000                      ' Used to test for SPI command codes
  i2cCmdMask  = %00001111                      ' and i2c bits, which test just the LSN
' Options for commands

  ioNoStore   = %00100000                      ' If set, data is not stored into main memory
                                               '  If ioBootCmd or ioSpiBoot, no cogs are
                                               '   stopped and a new cog is not started.
  ioLowSpeed  = %01000000                      ' If set, I2C runs at 100KHz rather than 400KHz
  ioStopLdr   = %10000000                      ' If set, the loader's cog is stopped after a boot

' Return status

  ioWriteErr  = %10000000                      ' An error occurred during an I2C write (NAK)

  ioTestRdy   = ioCmdMask << 24                ' Used to test 1st control long for ready
  ioTestErr   = ioWriteErr << 24               ' Used to test 1st control long for write error

OBJ
  ' temporary for debugging
  ' ideally we could fold serial into our IO system,
  ' but the nature of serial, timing, and lack of flow
  ' control may make that imposible
  ' for now... we haul it in and re-export it up for
  ' others to use
'  serial     : "FullDuplexserial_drv_014.spin"       ' the full duplex serial driver

PUB start(ctrlBlk) | spiConf[5], i2cConf[2]              '' Start the I2C I/O driver (standalone)
  ' this is called by the top level to launch the cog

  control := ctrlBlk                                   ''  using address of 2 longs for control
  stop                                                 ' Stop a previous copy

  computeTimes
'  clockCounterSetup := (constant(%00100 << 26) + PIN_SPI_SCK)     ' for high speed SPI mode
'  dataInCounterSetup := (constant(%00100 << 26) + PIN_SPI_MOSI)

  long[control][0] := 0                                ' Initialize the control block
  long[control][1] := 0
  long[control][2] := 42


  lockNumber := locknew + 1
  if lockNumber == 0
    return false

  cog := cognew(@Initialization,control) + 1            ' Start a new cog with the I2C driver
  if cog > 0                                       ' Indicate success
    ' do initial hardware init
    ' initialize SPI bus/mux
    doCommand(ioSpiInit, 0, @spiConf, 5)
    doCommand(i2cInit, 0, @i2cConf, 2)

    return (cog > 0)                                       ' Indicate success
  ' fell thru
  return false

PUB stop                                               '' Stop the I2C I/O driver (standalone)
  if cog > 0
    cogstop(cog - 1)

  if(lockNumber)
    lockret(-1 + lockNumber~)


PUB computeTimes                                       '' Set up timing constants in assembly
                                                       '  (Done this way to avoid overflow)
  i2cDataSet1 := ((clkfreq / 10000) *  900) / 100000   ' Data setup time -  900ns (100KHz)
  i2cClkLow1  := ((clkfreq / 10000) * 4700) / 100000   ' Clock low time  - 4700ns (100KHz)
  i2cClkHigh1 := ((clkfreq / 10000) * 4000) / 100000   ' Clock high time - 4000ns (100KHz)
  i2cDataSet4 := ((clkfreq / 10000) *  550) / 100000   ' Data setup time -  550ns (400KHz)
  i2cClkLow4  := ((clkfreq / 10000) * 1300) / 100000   ' Clock low time  - 1300ns (400KHz)
  i2cClkHigh4 := ((clkfreq / 10000) * 1000) / 100000   ' Clock high time - 1000ns (400KHz)
  ioPause    := clkfreq / 100000                       ' Pause between checks for operations


PUB getCog
  return cog
PUB getControl(i)                                      '' Return a long from the control block
  return long[control][i]                              ' Check for operation completed first

PUB setControl(i,value)                                '' Set value of a long in the control block
  long[control][i] := value                            ' Always set the first long last

PUB waitUntilReady
  repeat while long[control][0] & ioTestRdy            ' Wait for previous I/O to finish

PUB isError
  return (long[control][0] & ioTestErr)

PUB doCommand(cmd, addr, buffer, count ) | cmd0, cmd1
  cmd0 := (cmd << 24) | (addr & $FFFFFF)
  cmd1 := (count << 16) | (buffer & $FFFF)
  doCommandRaw(cmd0, cmd1)

PUB doCommandRaw(cmd0, cmd1)
  waitUntilReady                                    ' Wait for previous I/O to finish
  setControl(1, cmd1)
  setControl(0, cmd0)
  waitUntilReady                                    ' Wait for this to finish

PUB SPI_Select_Channel( channel )
  doCommand( ioSpiSelect, channel, 0, 0 )

PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:   num_bits : number of bits to transmit from data_out
         data_out : source of data to transmit
         bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits
RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := data_out
  doCommand( ioSpiWRBits, 0, @data_in, num_bits )
  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )
' end SPI_Write_Read

PUB SPI_Send_Recv_Byte(spi_data8) | spi_result8

' DESCRIPTION: this functions sends spi_data8 and receives spi_result8, remember all spi operations are circular, so
' when a byte is sent a byte is received at the same time, you can ignore the sent or received byte as you
' wish. For example, to do a read just send the dummy value $FF
' NOTE: assumes caller is controlling the chip select line
'
' INPUTS: spi_data8 - 8 bit data that is going to be sent
' OUTPUTS: returns the 8-bit data received at the same time
' -----------------------------------------------------------------------------

  ' reset result
  spi_result8 := spi_data8
  doCommand( ioSpiWriteRead, 0, @spi_result8, 1 )
  return (spi_result8)

PUB SPI_Read_Byte

' DESCRIPTION: reads a single byte from the SPI interface, simply a dummy call to SPI_Send_Recv_Byte
'
' INPUTS: none.
'
' OUTPUTS: returns the byte read from the SPI interface
' -----------------------------------------------------------------------------

  return ( SPI_Send_Recv_Byte($FF) )

' end SPI_Read_Byte

'//////////////////////////////////////////////////////////////////////////////

PUB SPI_Write_Byte(spi_data8)

' DESCRIPTION: sends a single byte to the SPI interface, returns the read value
'
' INPUTS: spi_data8 - 8-bit data to send
'
' OUTPUTS: returns the 8-bit data from the SPI interface
' -----------------------------------------------------------------------------

' return the data from the cylic buffer regardless...
return( SPI_Send_Recv_Byte(spi_data8) )

' end SPI_Write_Byte

PUB SPI_Write_Read_Bytes(buffer16, count) | spi_result8

' DESCRIPTION: this functions sends spi_data8 and receives spi_result8, remember all spi operations are circular, so
' when a byte is sent a byte is received at the same time, you can ignore the sent or received byte as you
' wish. For example, to do a read just send the dummy value $FF
' NOTE: assumes caller is controlling the chip select line
'
' INPUTS: spi_data8 - 8 bit data that is going to be sent
' OUTPUTS: returns the 8-bit data received at the same time
' -----------------------------------------------------------------------------
  doCommand( ioSpiWriteRead, 0, buffer16, count )

PUB SPI_Read_Bytes(buffer16, count)
  doCommand( ioSpiRead, 0, buffer16, count )

PUB SPI_Write_Bytes(buffer16, count)
  doCommand( ioSpiWrite, 0, buffer16, count )

PUB setLock ' 3 Stack Longs

  if(lockNumber)
    repeat while(lockset(lockNumber - 1))

PUB clearLock ' 3 Stack Longs

  if(lockNumber)
    lockclr(lockNumber - 1)


DAT

org 0

'' ######################################################################
''  IOCog runtime actions
'' ######################################################################

' This is both the top level entry point of the cog code, and the
' beginning of the loop that runs. There is no internal initialization
' to speak of really

'' The control information is
'' passed in a 2 long parameter block whose address is passed to the COG when it is started.
'' The parameter block is updated when the operation is completed.  Note that these are shown
'' here as they appear in a long value rather than the order of the bytes in memory.

'' -------------------------------------------------------------------
'' |   cmd/status   |          I/O pin / device / address            |
'' -------------------------------------------------------------------
'' |           byte count           |          HUB address           |
'' -------------------------------------------------------------------

' commands parsed as follows:
'    control[0] : high 8 bits  = command (i2cCmd)
'                 low 24 bits  = address for command (i2cAddr)
'    control[1] : high 16 bits = byte count (i2cCount)
'                 low 16 bits  = hub address of main buffer (i2cBufAddr)

Initialization
MainLoop                mov     ioTemp,ioPause       ' Wait 10us before checking
                        add     ioTemp,CNT
                        waitcnt ioTemp,#0

' processing of the command. communication happens via 2 longs
' in an array called 'control'. A pointer is passed not from IOCog
' but it's initialization owner to here. On with the show...
' first we procecess the first long, control[0]

NewOpFetch
                        rdlong  ioAddr,PAR             ' Fetch control information
                        mov     ioCmd,ioAddr
                        shr     ioCmd,#24              ' Isolate command code
                        mov     Options,ioCmd          ' not sure...
                        and     ioAddr,ioAddrMask     ' Only need address at this point

' if command byte is not valid command, we try again, looping to start
                        and     ioCmd,#ioCmdMask  wz
                if_z    jmp     #MainLoop          ' Wait for a new operation

' if it's a command, look to control[1], as ioCount here..
                        mov     ioTemp,PAR
                        add     ioTemp,#4              ' Now get 2nd long of packet
                        rdlong  ioCount,ioTemp        ' get unshifted count
                        mov     ioBufAdr,ioCount      ' Byte count
                        shr     ioCount,#16            ' shift down the count
                        and     ioBufAdr,ioWordMask   ' HUB RAM address of buffer
' send to the system-specific dispatchers.
' this should happen in reverse order to (or, most to least specific in terms of bit masks)
                        test    ioCmd,#spiCmdMask  wz   ' Check for SPI commands
                if_nz   jmp     #spiEntryPoint
                        test    ioCmd,#i2cCmdMask  wz   ' Check for I2C commands
                if_nz   jmp     #i2cEntryPoint
' no match, for now ignore, but FIXME is it should probably explicitly error?

                        ' fall through to end of command handler

'' The operation has completed, with or without errors.  Update the control block
'' in main memory and wait for the next operation to be requested.
'' Response reflects end state of what was passed in, save for clearing command
'' to be zero
'' This has many entry points, including falling in from above.
cmdGoUpdate             mov     ioTemp,PAR
                        add     ioTemp,#4
                        and     ioBufAdr,ioWordMask   ' Put back ioBufAddr
                        shl     ioCount,#16
                        or      ioCount,ioBufAdr      ' put back ioCount
                        wrlong  ioCount,ioTemp        ' write response control[1]
                        wrlong  ioAddr,PAR            ' write response control[0]
                        jmp     #MainLoop             '  and go wait for a new one
'' end of assembly loop

'' ######################################################################
''  SPI commands
'' ######################################################################

' the main command processor masks SPI above, this is the sub-processor
' this to make some bitwise assumptions about the command bits to
' optimize the flow
spiEntryPoint           cmp     ioCmd,#ioSpiSelect wc,wz    ' compare 0000000_
              if_c      jmp     #spiDoInit                  ' Decode operation
              if_z      jmp     #spiDoSelect
                        cmp     ioCmd,#ioSpiWrite wc,wz     ' compare 0000001_
              if_c      jmp     #spiDoRead
              if_z      jmp     #spiDoWrite
                        cmp     ioCmd,#ioSpiWRBits wc,wz    ' compare 0000010_
              if_c      jmp     #spiDoWriteRead
              if_z      jmp     #spiDoWRBits
                        cmp     ioCmd,#ioSpiWriteBits wc,wz ' compare 0000011_
              if_c      jmp     #spiDoReadBits
              if_z      jmp     #spiDoWriteBits
                        cmp     ioCmd,#ioSpiShutdown wc,wz  ' compare 0000100_
'              if_c      jmp     #cmdGoUpdate                ' no command found
              if_z      jmp     #spiDoStop
'                        or      i2cAddr,errorFlag
                        jmp     #cmdGoUpdate
' end of spiEntryPoint

'' ######################################################################
'' ######## COMMAND: ioSpiInit
'' ######################################################################

'' Initialize SPI communications.  The hub address field points to a table of
'' up to 6 longs that contain the SPI counter value for the SD card, the SPI
'' counter clear pin mask, the SPI counter count pin mask, the SPI MOSI pin
'' mask, the SPI MISO pin mask, and the SPI clock pin mask.  The count field
'' contains the number of longs provided.  Any not supplied remain unchanged
'' from the previous call.  The defaults are appropriate for the Prop C3 board.

' will return up to 6 bytes describing the configuration to the caller
' (this is probably not needed, and will be chucked at some point in the name
' of shrinkage

' EXPECTS: nothing
' RETURNS: 6 bytes of config info written to ioBufAdr

spiDoInit               movd    :copyTable,#hardwareConfig   ' get pointer to select counter
                        tjz     ioCount,#:skipCopy           ' skip on zero count
                        max     ioCount,#5                   ' Up to 5 table items
' puts the spi mask table into the return buffer
:copyTable              wrlong  0-0,ioBufAdr           ' not sure... seems to be copy
                        add     :copyTable,incrDst      ' adds 256 to the pointer?
                        add     ioBufAdr,#4            ' moves the i2cbuf ahead 1L
                        djnz    ioCount,#:copyTable    ' the 'until' on the copy

:skipCopy               or      dira,spiMaskClr        ' set direction registers
                        or      dira,spiMaskCnt
                        or      dira,spiMaskDI
                        or      dira,spiMaskClk
                        andn    outa,spiMaskCnt    'init some outputs
                        andn    outa,spiMaskClr    ' clear counter
                        or      outa,spiMaskClr    ' allow counting
                        andn    outa,spiMaskDI
                        andn    outa,spiMaskClk
                        jmp     #cmdGoUpdate


'' ######################################################################
'' ####### ACTION: ioSpiStop
'' ######################################################################

'' Stop SPI communications.  Any previously used I/O pins are set to input mode
'' FIXME: this should put the pins back to pre-init state, confirm...
'' EXPECTS: Nothing
'' RETURNS: Nothing

' forcibly returns to end of command, regardless of caller
spiDoStop               movs    spiDoStop_ret,#cmdGoUpdate
                        andn    outa,spiMaskClr         ' Make sure /CS is high
                        andn    dira,spiMaskClk         ' Set the card pins to
                        andn    dira,spiMaskDO          '  input mode but leave
spiDoStop_ret           ret

{
' Modification by Cluso99 to force the SD card to tristate the D0 pin (allows other chips on the bus)
spiDoStop               call    #deselectSD             ' CS=1
                        andn    dira,spiMaskDO          ' disable D0 as output (s/be input anyway)
                        call    #spiRecvByte            ' Output a stream of clocks
                        andn    dira,spiMaskClr         ' Clr input
                        'mov     spiMaskClr,#0          ' Fix this!!! Clr is not initialized
                        andn    dira,spiMaskInc         ' Inc input
                        mov     spiMaskInc,#0
                        mov     spiMaskDO,#0            ' D0 already input (above)
                        andn    dira,spiMaskDI          ' DI input
                        mov     spiMaskDI,#0            '
                        andn    dira,spiMaskClk         ' CLK input
                        mov     spiMaskClk,#0
                        jmp     #i2cGoUpdate
}

'' ######################################################################
'' ####### ACTION: ioSpiSelect
'' ######################################################################

' This will set the SPI selection to the value 0-7, passed in addr
' EXPECTS: that ioAddr is set to the address of the device to select
' RETURNS: Nothing

spiDoSelect             mov spiLayer1Param, ioAddr
                        call #spiSendSelect
SelectEnd               jmp     #cmdGoUpdate

'' ######################################################################
'' ####### ACTION: ioSpiRead
'' ####### ACTION: ioSpiWrite
'' ####### ACTION: ioSpiWriteRead
'' ######################################################################

' EXPECTS: ioCount is set to the number of bytes to read/write out
'          ioBufAdr points to the buffer of bytes to read/write on SPI
' RETURNS: the bytes asked for in ioBufAdr, if a read op

spiDoRead
spiDoWrite
spiDoWriteRead
                        tjz     ioCount,#writeReadEnd  ' Check for count exhausted
:writeReadRepeat        cmp     ioCmd, #ioSpiRead   wz ' figure out what to send
                 if_z   mov     localData, $FF         ' for read-only, send a $FF
                 if_nz  rdbyte  localData,ioBufAdr     ' for write ops, send the data in buffer
                        call    #spiSendRecvByte       ' send the byte
                        cmp     ioCmd, #ioSpiWrite  wz ' if the read isn't important to the op, don't
                 if_nz  wrbyte  localData,ioBufAdr     ' Store data in specified location
                        add     ioBufAdr,#1            ' increment the address pointer
                        djnz    ioCount,#:writeReadRepeat '  to next if more data remains
writeReadEnd            jmp     #cmdGoUpdate


'' ######################################################################
'' ####### ACTION: ioSpiReadBits
'' ####### ACTION: ioSpiWriteBits
'' ####### ACTION: ioSpiWRBits
'' ######################################################################

' EXPECTS: ioCount is set to the number of bits to read/write
'          ioBufAdr points a long of bits to read/write from/to SPI
' RETURNS: the bits asked for in the first long of ioBufAdr if a read op is implied

spiDoReadBits
spiDoWriteBits
spiDoWRBits             tjz     ioCount,#WRBitsEnd      ' Check for count exhausted
                        cmp     ioCmd, ioSpiReadBits wz ' determine what to write
                 if_z   mov     localData, #0           ' bits we care nothing about for readonly op
                 if_nz  rdlong  localData,ioBufAdr      ' get the long of bits for write ops
                        mov     localCount, ioCount
                        call    #spiSendRecvBits
                        cmp     ioCmd, ioSpiWriteBits wz ' Don't store read data on a write-only op
                 if_nz  wrlong  localData,ioBufAdr       ' Store data in specified location
WRBitsEnd               jmp     #cmdGoUpdate             ' we're done


'' ######################################################################
''  I2C commands
'' ######################################################################


'' ######################################################################
''  Above here: The higher level command level interface/actions
'' ----------------------------------------------------------------------
''  Below here: The lower level services used by the command level
'' ######################################################################


'' ######################################################################
''  SPI Low bit/byte level byte I/O
'' ######################################################################

'' expects a number of bits on localCount and
'' sends and receives via localData
spiRecvByte             mov localData, #$FF         '' blank bits we dont care about
spiSendRecvByte
spiSendByte             mov localCount, #8          ' we want a byte of transfer
spiSendRecvBits         max localCount, #32         ' at most 32 bits via this
                        mov spiBitsOut, localData   ' put data into the out buffer
                        mov spiBitsIn, #0           ' blank the in buffer
                        mov ioTemp, #32          ' figure how much to offset out
                        sub ioTemp, localCount
                        shl spiBitsOut, ioTemp   ' pad it to MSB
                        mov spiLayer1Param, localCount 'prepare call to do cycles
                        call #spiDoClockCycles            ' send the clocks
                        mov localData, spiBitsIn    ' set the return
spiSendRecvBits_ret
spiSendByte_ret
spiSendRecvByte_ret
spiRecvByte_ret         ret


'' ######################################################################
''  SPI Physical I/O (IE: lowest, bit level
'' ######################################################################

'' this is a very low level routine which shifts bits on and off the SPI
'' buffer, and in/out of spiBitsIn and spiBitsOut respectively
'' is it up up a higher (but still low) level routine to do the right thing
'' with these in/out buffers
'' this will take a counter, up to 32, for the number of cyckes


spiDoClockCycles        max     spiLayer1Param, #32   ' at most 32 bits to shift
                        tjz     spiLayer1Param,#spiDoClockCycles_ret ' skip if 0
:repeatclock            shl     spiBitsOut, #1   wc       ' shift off the bit to send
                        andn    outa,spiMaskClk       ' clock 0
                        muxc    outa,spiMaskDI        ' send bit
                        or      outa,spiMaskClk       ' clock 1
                        test    spiMaskDO,ina    wc ' get incoming bit
                        rcl     spiBitsIn,#1          ' and put it in buffer bit 1
                        djnz    spiLayer1Param, #:repeatClock
                        'or      outa,spiMaskDI          ' Leave DI in idle (high) state
                        andn    outa,spiMaskClk       ' clock 0
                        andn    outa,spiMaskDI       ' clock 0
spiDoClockCycles_ret    ret

'' the uses the C3's SPI muxer to select, or deselect a channel
spiSendSelect           max     spiLayer1Param, #7      ' loop is for 1-7, 0 comes before that
                        andn    outa,spiMaskClr         ' clear the 161 counter
                        or      outa,spiMaskClr         ' and clear the clear.
                        tjz     spiLayer1Param,#spiSendSelect_ret   ' Leave nothing selected if zero
                        andn    outa,spiMaskCnt         '  enable the right SPI select
:spiSelCnt              or      outa,spiMaskCnt         ' Output number of pulses needed to
                        andn    outa,spiMaskCnt         '
                        djnz    spiLayer1Param,#:spiSelCnt
spiSendSelect_ret       ret


'' ######################################################################
''  I2C commands
'' ######################################################################

'' i2c command dispatcher
i2cEntryPoint
                        movs    :getAction,ioCmd       ' Get command specific action mask
                        add     :getAction,#i2cActionTbl

                        test    Options,#ioLowSpeed wc  ' Set bus speed based on options
                if_c    mov     i2cDataSet,i2cDataSet1  ' Default to 100Khz speed
                if_c    mov     i2cClkLow,i2cClkLow1
                if_c    mov     i2cClkHigh,i2cClkHigh1
                if_nc   mov     i2cDataSet,i2cDataSet4  ' otherwise, 400Khz speed
                if_nc   mov     i2cClkLow,i2cClkLow4
                if_nc   mov     i2cClkHigh,i2cClkHigh4

                        cmp     ioCmd,#i2cInit    wz
                if_z    jmp     #i2cDoInit

:getAction              mov     Action,0-0
                        call    #i2cStart               ' Do a start sequence

'' ##### Device selection, and 1-2 byte 'addressing'

                        test    Action,#%000000001 wz
                if_z    jmp     #:skipAction0
'' Action: 0 - Write the EEPROM device select code for write mode and 2 address bytes
                        mov     localData,ioAddr         ' Construct a device select
                        shr     localData,#16             '  code for EEPROM write mode
                        mov     localMask,#%10000000
                        call    #i2cWrite               ' Send device select code
                if_c    jmp     #:doStop                ' Failure if NAK received

:skipAction0            test    Action,#%000000010 wz
                if_z    jmp     #:skipAction1
'' Action: 1 - Write the MSB device address or (for ioRead1Cmd/ioWrite1Cmd) a device select code
                        mov     localData,ioAddr         ' First address byte is most
                        shr     localData,#8              '  significant byte of address
                        mov     localMask,#%10000000
                        call    #i2cWrite               ' Send first address byte
                if_c    jmp     #:doStop                ' Failure if NAK received

:skipAction1            test    Action,#%000000100 wz
                if_z    jmp     #:skipAction2
'' Action: 2 - Write the LSB device address or (for ioRead0Cmd/ioWrite0Cmd) a device select code
                        mov     localData,ioAddr         ' Second address byte is least
                        mov     localMask,#%10000000      '  significant byte of address
                        call    #i2cWrite               ' Send second address byte
                if_c    jmp     #:doStop                ' Failure if NAK received

'' End of selection and 'addressing'

:skipAction2            tjz     ioCount,#:doStop       ' If byte count == 0, we're done

'' Action: 3 - Output a Start Sequence prior to reselecting in read mode after addressing
                        test    Action,#%000001000 wz
                if_nz   call    #i2cStart               ' Do a start sequence if readdressing

'' Action: 4 - Fetch a data value for writing
:doReadWrite            test    Action,#%000010000 wz
                if_nz   rdbyte  localData,ioBufAdr       ' If writing, fetch the data value
                if_nz   add     ioBufAdr,#1            '  and increment the hub address

                        test    Action,#%000100000 wz
                if_z    jmp     #:skipAction5
'' Action: 5 - Construct an EEPROM device select code for read mode and 2 address bytes
                        mov     localData,ioAddr         ' If reading, construct a device select
                        shr     localData,#16             '  code for EEPROM read mode with
                        or      localData,#%00000001      '   read mode given one for write mode

:skipAction5            test    Action,#%001000000 wz
                if_z    jmp     #:skipAction6
'' Action: 6 - Construct a read mode device select code from the MSB of the 16 bit device address
                        mov     localData,ioAddr         ' If reading using a single byte address
                        shr     localData,#8              '  construct a device select code for
                        or      localData,#%00000001      '   read mode given one for write mode

:skipAction6            test    Action,#%010000000 wz
                if_z    jmp     #:skipAction7
'' Action: 7 - Write the data value or read mode device select code
                        mov     localMask,#%10000000      ' Either readdress device for reading
                        call    #i2cWrite               '  or write a data value at this point
                if_c    jmp     #:doStop                ' Failure if NAK received


'' Action: 8 - Read a byte of data from the device and store it
:skipAction7            test    Action,#%100000000 wz
                if_z    jmp     #:skipAction8

                        cmp     ioCount,#2        wc   ' Carry true if this is the last byte
                        mov     localMask,#%10000000
                        mov     localData,#0
                        call    #i2cRead

                        wrbyte  localData,ioBufAdr       ' Store data in specified location
                        add     ioBufAdr,#1            '  and increment the address pointer

                        or      i2cZero,#0      nr,wc
'' loop for as many bytes as needed
                        andn    Action,#%011100000      ' No readdressing on subsequent reads
:skipAction8            add     ioAddr,#1
                        djnz    ioCount,#:doReadWrite  ' Repeat for number of bytes requested

'' finish command
:doStop                 call    #i2cStop
                if_c    or      ioAddr,errorFlag       ' Carry true indicates error
                        jmp     #cmdGoUpdate


'' ######################################################################
''  I2C Low bit/byte level byte I/O
'' ######################################################################

'' ######################################################################
'' ####### ACTION: i2cInit
'' ######################################################################

'' TODO: don't have inits for each system, combine them
i2cDoInit               movd    :copyTable,#i2cSCLMask ' get pointer to select counter
                        tjz     ioCount,#:skipCopy     ' skip on zero count
                        max     ioCount,#2             ' Up to 2 table items
:copyTable              wrlong  0-0,ioBufAdr           ' copy the masks to buffer
                        add     :copyTable,incrDst     ' next item to move
                        add     ioBufAdr,#4            ' next slot in buffer
                        djnz    ioCount,#:copyTable    ' until done
' now stop talking 'bout it, and DO something 'bout it.
:skipCopy               call    #i2cReset
                        ' TODO: move i2cReset into here... it's the only caller
                        jmp     #cmdGoUpdate


'' ######################################################################
''  I2C Low bit/byte level byte I/O
'' ######################################################################

'' Low level I2C routines.  These are designed to work either with a
'' standard I2C bus
'' (with pullups on both SCL and SDA) or the Propellor Demo Board (with a pullup only
'' on SDA).  Timing can be set by the caller to 100KHz or 400KHz.

'' ######################################################################
'' Do I2C Reset Sequence.  Clock up to 9 cycles.  Look for SDA high while SCL
'' is high.  Device should respond to next Start Sequence.  Leave SCL high.

i2cReset                andn    dira,i2cSDAMask         ' Pullup drive SDA high
                        mov     i2cBitCnt,#9            ' Number of clock cycles
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Allow for minimum SCL low
:i2cResetClk            andn    outa,i2cSCLMask         ' Active drive SCL low
                        or      dira,i2cSCLMask
                        waitcnt i2cTime,i2cClkHigh      ' For the minimum time
                        andn    dira,i2cSCLMask             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkLow       ' Allow minimum SCL high
                        test    i2cSDAMask,ina     wz   ' Stop if SDA is high
              if_z      djnz    i2cBitCnt,#:i2cResetClk ' Stop after 9 cycles
i2cReset_ret            ret                             ' Should be ready for Start


'' ######################################################################
'' Do I2C Start Sequence.  This assumes that SDA is a floating input and
'' SCL is also floating.
'' The start sequence is where SDA goes from HIGH to LOW while SCL is HIGH.

i2cStart                andn    dira,i2cSDAMask         ' Pullup drive SDA high
                        andn    outa,i2cSDAMask         ' SDA set to drive low

                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Allow for bus free time
                        waitcnt i2cTime,i2cClkHigh
                        andn    dira,i2cSCLMask             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkHigh      ' Allow for start setup time
                        or      dira,i2cSDAMask         ' Active drive SDA low
                        waitcnt i2cTime,#0              ' Allow for start hold time
                        andn    outa,i2cSCLMask         ' Active drive SCL low
                        or      dira,i2cSCLMask

i2cStart_ret            ret


'' ######################################################################
'' Do I2C Stop Sequence.  This assumes that SCL is low and SDA is indeterminant.
'' The stop sequence is where SDA goes from LOW to HIGH while SCL is HIGH.
'' i2cStart must have been called prior to calling this routine for initialization.
'' The state of the (c) flag is maintained so a write error can be reported.

i2cStop                 or      dira,i2cSDAMask         ' Active drive SDA low
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Wait for minimum clock low
                        waitcnt i2cTime,i2cClkLow
                        andn    dira,i2cSCLMask             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkHigh      ' Wait for minimum setup time
                        andn    dira,i2cSDAMask         ' Pullup drive SDA high
                        waitcnt i2cTime,#0              ' Allow for bus free time
                        andn    dira,i2cSCLMask         ' Leave SCL and SDA high
i2cStop_ret             ret


'' ######################################################################
'' Write I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant. The (c) flag will be set on exit from ACK/NAK with ACK == false
'' and NAK == true. Bytes are handled in "little-endian" order so these routines can be
'' used with words or longs although the bits are in msb..lsb order.

i2cWrite                mov     i2cBitCnt,#8
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Wait for minimum SCL low
:i2cWriteBit            waitcnt i2cTime,i2cDataSet
                        test    localData,localMask    wz
              if_z      or      dira,i2cSDAMask         ' Copy data bit to SDA
              if_nz     andn    dira,i2cSDAMask
                        waitcnt i2cTime,i2cClkHigh      ' Wait for minimum setup time
                        andn    dira,i2cSCLMask             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkLow
                        andn    outa,i2cSCLMask         ' Active drive SCL low
                        or      dira,i2cSCLMask
                        ror     localMask,#1            ' Go do next bit if not done
                        djnz    i2cBitCnt,#:i2cWriteBit

                        andn    dira,i2cSDAMask         ' Switch SDA to input and

                        waitcnt i2cTime,i2cClkHigh      '  wait for minimum SCL low

                        andn    dira,i2cSCLMask             ' Pullup drive SCL high

                        waitcnt i2cTime,#0              ' Wait for minimum high time

                        test    i2cSDAMask,ina     wc   ' Sample SDA (ACK/NAK) then

' Whatthafaack?
                        nop
                        nop
                        nop

                        andn    outa,i2cSCLMask         '  active drive SCL low
                        or      dira,i2cSCLMask
                        or      dira,i2cSDAMask         ' Leave SDA low

                        rol     localMask,#16           ' Prepare for multibyte write
i2cWrite_ret            ret

'' ######################################################################
'' Read I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant.  ACK/NAK will be copied from the (c) flag on entry with
'' ACK == low and NAK == high.  Bytes are handled in "little-endian" order so these
'' routines can be used with words or longs although the bits are in msb..lsb order.

i2cRead                 mov     i2cBitCnt,#8
                        andn    dira,i2cSDAMask         ' Make sure SDA is set to input
                        mov     i2cTime,i2cClkLow
                        add     i2cTime,cnt             ' Wait for minimum SCL low
:i2cReadBit             waitcnt i2cTime,i2cClkHigh
                        andn    dira,i2cSCLMask             ' Pullup drive SCL high
                        waitcnt i2cTime,i2cClkLow       ' Wait for minimum clock high
                        test    i2cSDAMask,ina     wz   ' Sample SDA for data bits

                        andn    outa,i2cSCLMask         ' Active drive SCL low
                        or      dira,i2cSCLMask

              if_nz     or      localData,localMask     ' Accumulate data bits
              if_z      andn    localData,localMask
                        ror     localMask,#1              ' Shift the bit mask and
                        djnz    i2cBitCnt,#:i2cReadBit  '  continue until done

                        waitcnt i2cTime,i2cDataSet      ' Wait for end of SCL low

              if_c      andn    dira,i2cSDAMask         ' Copy the ACK/NAK bit to SDA
              if_nc     or      dira,i2cSDAMask
                        waitcnt i2cTime,i2cClkHigh      ' Wait for minimum setup time
                        andn    dira,i2cSCLMask             ' Pullup drive SCL high
                        waitcnt i2cTime,#0              ' Wait for minimum clock high

                        andn    outa,i2cSCLMask         ' Active drive SCL low
                        or      dira,i2cSCLMask

                        or      dira,i2cSDAMask         ' Leave SDA low
                        rol     localMask,#16           ' Prepare for multibyte read
i2cRead_ret             ret


'' ######################################################################
''  Variables and constants, end of code
'' ######################################################################

'' global vars, previously in VAR, but that would not share well, so
'' trying to move it here...

cog                     long    0
lockNumber              long    0 ' Driver lock number.

control                 long    0

'' Constants for all routines
ioAddrMask              long    $00FFFFFF
ioWordMask              long    $0000FFFF
incrDst                 long    1 << 9                  ' Used to increment destination field
errorFlag               long    $80000000               ' NAK received during write cycle

'' Variables for all routines

ioPause                 long    0                       ' Pause before re-fetching next operation
ioTemp                  long    0                       ' general temp use
ioCmd                   long    0                       ' the command being processed
ioAddr                  long    0                       ' the 'address' of the command
ioCount                 long    0                       ' the byte count for the command
ioBufAdr                long    0                       ' the buffer address for the command
Options                 long    0                       ' Option bits (ioNoStore, ioLowSpeed)

'' Local variables for low level routines

localMask               long    0                       ' Bit mask for bit to be tx / rx
localData               long    0                       ' Data to be transmitted / received
localCount              long    0                       ' Counter used for various purposes

'' Variables for I2C routines

i2cClkHigh              long    0                       ' Minimum clock high time (ticks)
i2cDataSet1             long    0                       ' Minumum data setup time (ticks) 100KHz
i2cClkLow1              long    0                       ' Minimum clock low time (ticks) 100KHz
i2cClkHigh1             long    0                       ' Minimum clock high time (ticks) 100KHz
i2cDataSet4             long    0                       ' Minumum data setup time (ticks) 400KHz
i2cClkLow4              long    0                       ' Minimum clock low time (ticks) 400KHz
i2cClkHigh4             long    0                       ' Minimum clock high time (ticks) 400KHz
i2cTime                 long    0                       ' Used for timekeeping
i2cClkLow               long    0                       ' Minimum clock low time (ticks)
i2cDataSet              long    0                       ' Minumum data setup time (ticks)
i2cAction               long    0


'' This action table contains bit sequences for controlling device addressing and read/write
'' mode selection for each of the commands possible.  From LSB to MSB, the actions are:
'' 0 - Write the EEPROM device select code for write mode and 2 address bytes
'' 1 - Write the MSB device address or (for ioRead1Cmd/ioWrite1Cmd) a device select code
'' 2 - Write the LSB device address or (for ioRead0Cmd/ioWrite0Cmd) a device select code
'' 3 - Output a Start Sequence prior to reselecting in read mode
'' 4 - Fetch a data value for writing
'' 5 - Construct an EEPROM device select code for read mode and 2 address bytes
'' 6 - Construct a read mode device select code from the MSB of the 16 bit device address
'' 7 - Write the data value or read mode device select code
'' 8 - Read a byte of data from the device and store it
i2cZero
i2cActionTbl            long    %0000000000             ' Command not used (indicates done)
                        long    %0110101111,%0010010111 ' Read/Write with 2 bytes of addressing
                        long    %0111001110,%0010010110 ' Read/Write with 1 byte of addressing
                        long    %0100000100,%0010010100 ' Read/Write data only
Action                  long    0


'' i2c only stuff that chould be pushed "up"
i2cBitCnt               long    0                       ' Number of bits to tx / rx

'' Variables for SPI routines
'' "Layer 1" physical IO

spiBitsIn               long    0      ' buffer used for incoming bits
spiBitsOut              long    0      ' buffer used for outgoing bits
spiLayer1Param          long    0      ' counter used only in very low level


'' NOTE: code depends on specific order from here to ENDNOTE
'' masks for SPI pin manipulation
hardwareConfig
spiMaskClr              long    1<<PIN_SPI_SEL_CLR      ' Mask for SPI select counter clear (25)
spiMaskCnt              long    1<<PIN_SPI_SEL_CLK       ' Mask for SPI select counter count ( 8)
spiMaskDI               long    1<<PIN_SPI_MOSI       ' Mask for SPI MOSI (DI) pin        ( 9)
spiMaskDO               long    1<<PIN_SPI_MISO      ' Mask for SPI MISO (DO) pin        (10)
spiMaskClk              long    1<<PIN_SPI_SCK      ' Mask for SPI clock pin            (11)
                        nop
'                        nop
                        nop
'' masks for i2C pin manipulation
i2cSCLMask              long    1<<28                       ' Bit mask for SCL
i2cSDAMask              long    1<<29                       ' Bit mask for SDA
'' ENDNOTE

'' END OF FILE (above should stay at end)
''
                        fit   496


{{
                            TERMS OF USE: MIT License

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}
'//////////////////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////


